#include <stdio.h>
#include <stdio.h>
#include "../annotation/annot_region_tools.h"
#include "../annotation/gff_utils.h"
#include "gencode_gtf.h"
#include "../../utils/file/utils.h"
#include "../../../lib/utils/ansi_string/ansi_string.h"
#include <algorithm>

char* get_entry(t_string_tokens* line_str_tokens, char* entry_id);

void parse_GENCODE_gff_grp_strs(vector<t_annot_region*>* all_regions)
{
	for(int i_r = 0; i_r < all_regions->size(); i_r++)
	{
		if(i_r % 1000 == 0)
		{
			fprintf(stderr, "Parsing grp string: %d (%d)      \r", i_r, all_regions->size());
		}

		t_gff_info* cur_gff_info = (t_gff_info*)(all_regions->at(i_r)->data);
		char* grp_str = cur_gff_info->group_str;

		// Allocate the property id's and values for this entry.
		cur_gff_info->prop_ids = new vector<char*>();
		cur_gff_info->prop_vals = new vector<char*>();

		// Parse the parent group string.
		t_string* grp_str_str = new t_string(grp_str);
		t_string_tokens* grp_str_tokens = grp_str_str->tokenize_by_chars(";\" ");

		int i = 0; 
		while(i < grp_str_tokens->size())
		{
			if(i+1 >= grp_str_tokens->size())
			{
				fprintf(stderr, "Could not parse the GENCODE gff group string:\n%s\n", grp_str);
				exit(0);
			}

			//fprintf(stderr, "Prop %s: %s\n", grp_str_tokens->at(i)->str(), grp_str_tokens->at(i+1)->str());

			cur_gff_info->prop_ids->push_back(t_string::copy_me_str(grp_str_tokens->at(i)->str()));
			i++;
			cur_gff_info->prop_vals->push_back(t_string::copy_me_str(grp_str_tokens->at(i)->str()));
			i++;
		} // i loop.

		t_string::clean_tokens(grp_str_tokens);
		delete grp_str_str;
	} // i_r loop.
}

t_Gencode_GTF::t_Gencode_GTF(char* gencode_gtf_file)
{
	this->entries = new vector<t_gencode_entry*>();

	FILE* f_gtf = open_f(gencode_gtf_file, "r");

	// Read the file.
	while(1)
	{
		if(entries->size() % 10000 == 0)
		{
			fprintf(stderr, "%d               \r", entries->size());
		}

		char* cur_line = getline(f_gtf);

		if(cur_line == NULL)
		{
			break;
		}

		if(cur_line[0] != '#')
		{
			//printf("Processing %s\n", cur_line);
			//getc(stdin);

			// Parse the line and get the entries.
			char _chrom[100];
			char _source[100];
			char _feature[100];
			int start;
			int end;
			char strand;

			t_string* cur_line_str = new t_string(cur_line);

			// Parse all the entries.
			t_string_tokens* line_tokens = cur_line_str->tokenize_by_chars(" \t");

			strcpy(_chrom, line_tokens->at(0)->str());
			strcpy(_source, line_tokens->at(1)->str());
			strcpy(_feature, line_tokens->at(2)->str());
			start = atoi(line_tokens->at(3)->str());
			end = atoi(line_tokens->at(4)->str());
			strand = (line_tokens->at(6)->str())[0];

			// Sanity check on the strand character.
			if(strand != '-' &&
				strand != '+')
			{
				printf("Strand char unrecognized: %c out of %s\n", strand, line_tokens->at(6)->str());
				exit(0);
			}

			// Resolve the id's, status's, type's.
			char chrom = resolve_chromosome(_chrom);
			char source = resolve_source(_source);
			char feature = resolve_feature(_feature);

			// Get entry allocates, dont reallocate the corresponding entries.
			char* gene_id = get_entry(line_tokens, "gene_id");
			char* _gene_status = get_entry(line_tokens, "gene_status");
			char* _gene_type = get_entry(line_tokens, "gene_type");
			char* gene_name = get_entry(line_tokens, "gene_name");
			char gene_status = resolve_status(_gene_status);
			char gene_type = resolve_biotype(_gene_type);

			char* transcript_id = get_entry(line_tokens, "transcript_id");
			char* _transcript_status = get_entry(line_tokens, "transcript_status");
			char* _transcript_type = get_entry(line_tokens, "transcript_type");
			char* transcript_name = get_entry(line_tokens, "transcript_name");
			char transcript_status = resolve_status(_transcript_status);
			char transcript_type = resolve_biotype(_transcript_type);

			// Resolve the status, and type for gene and transcript.

			// Allocate the new entry and add it to the list: Resolve id's copy indices, etc.
			/*
			char chr_id; // From enumerated id's.
			char feature_id; // From enumerated id's.
			char source_id; // From enumerated id's.
			int start;
			int end;
			char strand;

			char gene_type; // From enumerated biotypes.
			char* gene_id;
			char* gene_name;
			char gene_status;
			char* transcript_id;
			char transcript_type; // From enumerated biotypes.
			char* transcript_name;
			char transcript_status;

			char status;
			*/
			// Validate all the values.
			if(chrom < N_CHRS && 
				source < N_SOURCES &&
				feature < N_FEATURES &&
				gene_type < N_BIOTYPES &&
				transcript_type < N_BIOTYPES && 
				gene_status < N_STATUS &&
				transcript_status < N_STATUS)
			{
				t_gencode_entry* new_entry = new t_gencode_entry();
				//new_entry->chr_id = resolve_chr_id(chrom);
				new_entry->chr_id = chrom;
				new_entry->source_id = source;
				new_entry->feature_id = feature;
				new_entry->start = start;
				new_entry->end = end;
				new_entry->strand = strand;

				new_entry->gene_type = gene_type;
				new_entry->gene_id = gene_id;
				new_entry->gene_name = gene_name;
				new_entry->gene_status = gene_status;
				new_entry->transcript_type = transcript_type;
				new_entry->transcript_id = transcript_id;
				new_entry->transcript_name = transcript_name;
				new_entry->transcript_status = transcript_status;

				//printf("%s:\n", cur_line);
				//printf("feature: %s\n", features[feature]);
				//printf("source: %s\n", sources[source]);
				//printf("start-end: %d-%d\n", start, end);
				//printf("gene_id: %s\n", gene_id);
				//printf("gene_type: %s\n", biotypes[gene_type]);
				//printf("transcript_type: %s\n", biotypes[transcript_type]);
				//printf("-------------------------------------\n");
				//getc(stdin);

				// Add the new entry.
				entries->push_back(new_entry);
			}

			// Clean the tokens.
			cur_line_str->clean_tokens(line_tokens);
			delete(cur_line_str);
		}
	} // Main file reading loop.

	fclose(f_gtf);

	printf("Loaded %d entries.\n", entries->size());
}

char resolve_status(char* status)
{
	for(int i_st = 0; i_st < N_STATUS; i_st++)
	{
		if(t_string::compare_strings_ci(statuses[i_st], status))
		{
			return(i_st);
		}
	} // i_st loop.

	printf("Could not resolve %s @ %s(%d)\n", status, __FILE__, __LINE__);
	return(N_STATUS);
}

char resolve_source(char* source)
{
	for(int i_st = 0; i_st < N_SOURCES; i_st++)
	{
		if(t_string::compare_strings_ci(sources[i_st], source))
		{
			return(i_st);
		}
	} // i_st loop.

	printf("Could not resolve @ %s(%d)\n", __FILE__, __LINE__);
	return(N_SOURCES);
}

char resolve_chromosome(char* chrom)
{
	for(int i_st = 0; i_st < N_CHRS; i_st++)
	{
		if(t_string::compare_strings_ci(chromosomes[i_st], chrom))
		{
			return(i_st);
		}
	} // i_st loop.

	printf("Could not resolve %s @ %s(%d)\n", chrom, __FILE__, __LINE__);
	return(N_CHRS);
}

char resolve_feature(char* feature)
{
	for(int i_st = 0; i_st < N_FEATURES; i_st++)
	{
		if(t_string::compare_strings_ci(features[i_st], feature))
		{
			return(i_st);
		}
	} // i_st loop.

	printf("Could not resolve %s @ %s(%d)\n", feature, __FILE__, __LINE__);
	return(N_FEATURES);
}

char resolve_biotype(char* type)
{
	for(int i_st = 0; i_st < N_BIOTYPES; i_st++)
	{
		if(t_string::compare_strings_ci(biotypes[i_st], type))
		{
			return(i_st);
		}
	} // i_st loop.

	printf("Could not resolve %s @ %s(%d)\n", type, __FILE__, __LINE__);
	return(N_BIOTYPES);
}

char* get_entry(t_string_tokens* line_str_tokens, char* entry_id)
{
	char* entry_str = NULL;
	for(int i_token = 0; i_token < line_str_tokens->size(); i_token++)
	{
		if(t_string::compare_strings_ci(entry_id, line_str_tokens->at(i_token)->str()))
		{
			entry_str = new char[strlen(line_str_tokens->at(i_token+1)->str()) + 2];
			strcpy(entry_str, line_str_tokens->at(i_token+1)->str());

			// Following is necessary to get rid of extra characters around the string.
			t_string::replace_avoid_list(entry_str, ";\"", 0);
		}
	} // i_token loop.

	return(entry_str);
}

t_Gencode_GTF::~t_Gencode_GTF()
{
}

vector<t_annot_region*>* get_introns()
{
	return(NULL);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_feature(vector<t_gencode_entry*>* entries, char* feature)
{
	vector<t_gencode_entry*>* entries_by_feature = new vector<t_gencode_entry*>();

	int feature_id_2_search = resolve_feature(feature);

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		if(entries->at(i_ent)->feature_id == feature_id_2_search)
		{
			entries_by_feature->push_back(entries->at(i_ent));
		}
	} // i_ent

	return(entries_by_feature);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_strand(vector<t_gencode_entry*>* entries, char strand)
{
	vector<t_gencode_entry*>* entries_by_feature = new vector<t_gencode_entry*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		if(entries->at(i_ent)->strand == strand)
		{
			entries_by_feature->push_back(entries->at(i_ent));
		}
	} // i_ent

	return(entries_by_feature);
}

vector<t_annot_region*>* t_Gencode_GTF::get_gtf_annot_regions_by_strand(vector<t_gencode_entry*>* entries, char strand)
{
	vector<t_annot_region*>* regions_by_strand = new vector<t_annot_region*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		if(entries->at(i_ent)->strand == strand)
		{
			//entries_by_feature->push_back(entries->at(i_ent));
			t_annot_region* new_region = new t_annot_region();
			new_region->chrom = new char[strlen(chromosomes[entries->at(i_ent)->chr_id]) + 2];
			strcpy(new_region->chrom, chromosomes[entries->at(i_ent)->chr_id]);
			new_region->start = entries->at(i_ent)->start;
			new_region->end = entries->at(i_ent)->end;
			new_region->strand = entries->at(i_ent)->strand;

			regions_by_strand->push_back(new_region);
		}
	} // i_ent

	return(regions_by_strand);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_gene_biotypes(vector<t_gencode_entry*>* entries, 
																			vector<int>* biotypes)
{
	vector<t_gencode_entry*>* entries_by_biotypes = new vector<t_gencode_entry*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		bool cur_enty_matches = false;
		for(int i_type = 0; 
				!cur_enty_matches && i_type < biotypes->size(); 
				i_type++)
		{
			if(entries->at(i_ent)->gene_type == biotypes->at(i_type))
			{
				entries_by_biotypes->push_back(entries->at(i_ent));
				cur_enty_matches = true;
			}
		} // i_type loop.
	} // i_ent

	return(entries_by_biotypes);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_transcript_biotypes(vector<t_gencode_entry*>* entries, 
																				vector<int>* biotypes)
{
	vector<t_gencode_entry*>* entries_by_biotypes = new vector<t_gencode_entry*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		bool cur_enty_matches = false;
		for(int i_type = 0; 
				!cur_enty_matches && i_type < biotypes->size(); 
				i_type++)
		{
			if(entries->at(i_ent)->transcript_type == biotypes->at(i_type))
			{
				entries_by_biotypes->push_back(entries->at(i_ent));
				cur_enty_matches = true;
			}
		} // i_type loop.
	} // i_ent

	return(entries_by_biotypes);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_chr_id(vector<t_gencode_entry*>* entries, int chr_id)
{
	vector<t_gencode_entry*>* entries_by_chr_id = new vector<t_gencode_entry*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		if(entries->at(i_ent)->chr_id == chr_id)
		{
			entries_by_chr_id->push_back(entries->at(i_ent));
		}
	} // i_ent

	return(entries_by_chr_id);
}

vector<t_gencode_entry*>* t_Gencode_GTF::get_gtf_entries_by_feature(char* feature)
{
	vector<t_gencode_entry*>* entries_by_feature = new vector<t_gencode_entry*>();

	int feature_id_2_search = resolve_feature(feature);

	for(int i_ent = 0; i_ent < this->entries->size(); i_ent++)
	{
		if(this->entries->at(i_ent)->feature_id == feature_id_2_search)
		{
			entries_by_feature->push_back(this->entries->at(i_ent));
			//printf("Added %d-%d\n", this->entries->at(i_ent)->start, this->entries->at(i_ent)->end);
			//getc(stdin);
		}
	} // i_ent

	return(entries_by_feature);
}

void t_Gencode_GTF::prune_entries_by_start_end(vector<t_gencode_entry*>* entries)
{
	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		t_gencode_entry* cur_gencode_entry = entries->at(i_ent);
		for(int i_ent2 = i_ent+1; i_ent2 < entries->size(); i_ent2++)
		{
			if(cur_gencode_entry->start == entries->at(i_ent2)->start &&
				cur_gencode_entry->end == entries->at(i_ent2)->end &&
				cur_gencode_entry->strand == entries->at(i_ent2)->strand &&
				cur_gencode_entry->chr_id == entries->at(i_ent2)->chr_id)
			{
				printf("Pruning %d. entry.\n", i_ent2);
				getc(stdin);

				// Prune this entry at i_ent2.
				entries->erase(entries->begin() + i_ent2);

				// Move i_ent2 back.
				i_ent2--;
			}
		}
	} // i_ent loop.
}

vector<t_annot_region*>* t_Gencode_GTF::get_processed_transcript_regions()
{
	vector<t_annot_region*>* processed_transcript_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<int>* p_t_biotypes = new vector<int>();
	p_t_biotypes->push_back(processed_transcript);
	vector<t_gencode_entry*>* processed_transcripts = t_Gencode_GTF::get_gtf_entries_by_transcript_biotypes(this->entries, 
																											p_t_biotypes);

	//t_Gencode_GTF::prune_entries_by_start_end(genes);

	printf("%d processed_transcript entries.\n", processed_transcripts->size());

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < processed_transcripts->size(); i_reg++)
	{
		t_annot_region* cur_p_t_region = new t_annot_region();
		cur_p_t_region->strand = processed_transcripts->at(i_reg)->strand;
		cur_p_t_region->start = processed_transcripts->at(i_reg)->start;
		cur_p_t_region->end = processed_transcripts->at(i_reg)->end;
		cur_p_t_region->chrom = new char[10];
		strcpy(cur_p_t_region->chrom, chromosomes[processed_transcripts->at(i_reg)->chr_id]);

		processed_transcript_regions->push_back(cur_p_t_region);
	} // i_reg loop.

	return(processed_transcript_regions);
}

vector<t_annot_region*>* t_Gencode_GTF::get_regions_per_encode_entries(vector<t_gencode_entry*>* entries)
{
	vector<t_annot_region*>* regions = new vector<t_annot_region*>();

	for(int i_ent = 0; i_ent < entries->size(); i_ent++)
	{
		t_annot_region* new_region = new t_annot_region();
		new_region->chrom = new char[strlen(chromosomes[entries->at(i_ent)->chr_id] + 2)];
		strcpy(new_region->chrom, chromosomes[entries->at(i_ent)->chr_id]);

		new_region->start = entries->at(i_ent)->start;
		new_region->end = entries->at(i_ent)->end;
		new_region->strand = entries->at(i_ent)->strand;
		regions->push_back(new_region);
	} // i_ent loop.

	return(regions);
}

// Get the regions between all the genes.
vector<t_annot_region*>* t_Gencode_GTF::get_intergenic_regions()
{
	vector<t_annot_region*>* intergenic_regions = new vector<t_annot_region*>();

	vector<t_gencode_entry*>* genes = this->get_gtf_entries_by_feature("gene");

	printf("%d gene entries.\n", genes->size());

	// Go over all the chromosomes.
	for(int i_chr = 0; i_chr < N_CHRS; i_chr++)
	{
		vector<t_gencode_entry*>* cur_chr_genes = t_Gencode_GTF::get_gtf_entries_by_chr_id(genes, i_chr);

		//vector<t_annot_region*>* plus_genes = t_Gencode_GTF::get_gtf_annot_regions_by_strand(cur_chr_genes, '+');
		//vector<t_annot_region*>* neg_genes = t_Gencode_GTF::get_gtf_annot_regions_by_strand(cur_chr_genes, '-');
		//printf("%d (+)gene entries on %s.\n", plus_genes->size(), chromosomes[i_chr]);
		//printf("%d (-)gene entries on %s.\n", neg_genes->size(), chromosomes[i_chr]);
		vector<t_annot_region*>* gene_regions = t_Gencode_GTF::get_regions_per_encode_entries(cur_chr_genes);

		// This is an important part of the generating the intergenic regions since there can be overlapping genes in the gencode annotation coming from HAVANA/ENSEMBL sources.
		//vector<t_annot_region*>* merged_plus_genes = merge_annot_regions(plus_genes, 0);
		//vector<t_annot_region*>* merged_neg_genes = merge_annot_regions(neg_genes, 0);
		vector<t_annot_region*>* merged_gene_regions = merge_annot_regions(gene_regions, 0);

		//delete_annot_regions(plus_genes);
		//delete_annot_regions(neg_genes);
		delete_annot_regions(gene_regions);

		//plus_genes = merged_plus_genes;
		//neg_genes = merged_neg_genes;
		gene_regions = merged_gene_regions;

		// Sort the genes wih respect to the coordinates.
		//sort(plus_genes->begin(), plus_genes->end(), sort_regions);
		//sort(neg_genes->begin(), neg_genes->end(), sort_regions);
		sort(gene_regions->begin(), gene_regions->end(), sort_regions);

		// For each intergenic region, dump the information.
		//for(int i_gene = 1; i_gene < plus_genes->size(); i_gene++)
		for(int i_gene = 1; i_gene < gene_regions->size(); i_gene++)
		{
			// Ensure that there is no overlap.
			//if(plus_genes->at(i_gene-1)->end+200 < plus_genes->at(i_gene)->start-200)
			if(gene_regions->at(i_gene-1)->end+200 < gene_regions->at(i_gene)->start-200)
			{
				//t_annot_region* cur_intergenic_region = new t_annot_region();
				//cur_intergenic_region->strand = '+';
				//cur_intergenic_region->start = plus_genes->at(i_gene-1)->end+200;
				//cur_intergenic_region->end = plus_genes->at(i_gene)->start-200;
				//cur_intergenic_region->chrom = new char[10];
				//strcpy(cur_intergenic_region->chrom, plus_genes->at(i_gene)->chrom);
				//intergenic_regions->push_back(cur_intergenic_region);

				t_annot_region* cur_intergenic_region = new t_annot_region();
				cur_intergenic_region->strand = '+';
				cur_intergenic_region->start = gene_regions->at(i_gene-1)->end+200;
				cur_intergenic_region->end = gene_regions->at(i_gene)->start-200;
				cur_intergenic_region->chrom = new char[10];
				strcpy(cur_intergenic_region->chrom, gene_regions->at(i_gene)->chrom);
				intergenic_regions->push_back(cur_intergenic_region);

				//printf("Added <%d-%d>\n", cur_intergenic_region->start, cur_intergenic_region->end);
			}
			else
			{
				//printf("Overlap: <%d-%d>, <%d-%d>\n", 
				//	plus_genes->at(i_gene-1)->start, plus_genes->at(i_gene-1)->end,
				//	plus_genes->at(i_gene)->start, plus_genes->at(i_gene)->end);

				//getc(stdin);
			}
		} // i_gene loop.

		//for(int i_gene = 1; i_gene < neg_genes->size(); i_gene++)
		//{
		//	if(neg_genes->at(i_gene-1)->end+200 < neg_genes->at(i_gene)->start-200)
		//	{
		//		t_annot_region* cur_intergenic_region = new t_annot_region();
		//		cur_intergenic_region->strand = '-';
		//		cur_intergenic_region->start = neg_genes->at(i_gene-1)->end+200;
		//		cur_intergenic_region->end = neg_genes->at(i_gene)->start-200;
		//		cur_intergenic_region->chrom = new char[10];
		//		strcpy(cur_intergenic_region->chrom, neg_genes->at(i_gene)->chrom);

		//		intergenic_regions->push_back(cur_intergenic_region);
		//		printf("Added <%d-%d>\n", cur_intergenic_region->start, cur_intergenic_region->end);		
		//	}
		//	else
		//	{
		//		//printf("Overlap: <%d-%d>, <%d-%d>\n", 
		//		//	neg_genes->at(i_gene-1)->start, neg_genes->at(i_gene-1)->end,
		//		//	neg_genes->at(i_gene)->start, neg_genes->at(i_gene)->end);

		//		//getc(stdin);
		//	}
		//} // i_gene loop.

		//cur_chr_genes->clear();
		//plus_genes->clear();
		//neg_genes->clear();
		//delete_annot_regions(cur_chr_genes);
		delete_annot_regions(gene_regions);
		//delete(plus_genes);
		//delete(neg_genes);
	} // i_chr loop.

	// Return the regions between the genes.
	return(intergenic_regions);
}

vector<t_annot_region*>* t_Gencode_GTF::get_CDS_regions()
{
	vector<t_annot_region*>* CDS_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<t_gencode_entry*>* CDSs = this->get_gtf_entries_by_feature("CDS");

	//t_Gencode_GTF::prune_entries_by_start_end(genes);

	printf("%d CDS entries.\n", CDSs->size());

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < CDSs->size(); i_reg++)
	{
		t_annot_region* cur_CDS_region = new t_annot_region();
		cur_CDS_region->strand = CDSs->at(i_reg)->strand;
		cur_CDS_region->start = CDSs->at(i_reg)->start;
		cur_CDS_region->end = CDSs->at(i_reg)->end;
		cur_CDS_region->chrom = new char[10];
		strcpy(cur_CDS_region->chrom, chromosomes[CDSs->at(i_reg)->chr_id]);

		CDS_regions->push_back(cur_CDS_region);
	} // i_reg loop.

	return(CDS_regions);
}


vector<t_annot_region*>* t_Gencode_GTF::get_UTR_regions()
{
	vector<t_annot_region*>* UTR_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<t_gencode_entry*>* UTRs = this->get_gtf_entries_by_feature("UTR");

	//t_Gencode_GTF::prune_entries_by_start_end(genes);

	printf("%d UTR entries.\n", UTRs->size());

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < UTRs->size(); i_reg++)
	{
		t_annot_region* cur_UTR_region = new t_annot_region();
		cur_UTR_region->strand = UTRs->at(i_reg)->strand;
		cur_UTR_region->start = UTRs->at(i_reg)->start;
		cur_UTR_region->end = UTRs->at(i_reg)->end;
		cur_UTR_region->chrom = new char[10];
		strcpy(cur_UTR_region->chrom, chromosomes[UTRs->at(i_reg)->chr_id]);

		UTR_regions->push_back(cur_UTR_region);
	} // i_reg loop.

	return(UTR_regions);
}

vector<t_annot_region*>* t_Gencode_GTF::get_ncRNA_regions_by_type(int ncrna_type)
{
	// Do filtering on the correct biotypes.
	vector<t_annot_region*>* ncRNA_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<t_gencode_entry*>* genes = this->get_gtf_entries_by_feature("gene");

	vector<int>* ncrna_biotypes = new vector<int>();

	ncrna_biotypes->push_back(ncrna_type);
	vector<t_gencode_entry*>* ncRNAs = this->get_gtf_entries_by_gene_biotypes(genes, ncrna_biotypes);

	//t_Gencode_GTF::prune_entries_by_start_end(genes);

	printf("%d %s entries.\n", ncRNAs->size(), biotypes[ncrna_type]);

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < ncRNAs->size(); i_reg++)
	{
		t_annot_region* cur_ncrna_region = new t_annot_region();
		cur_ncrna_region->strand = ncRNAs->at(i_reg)->strand;
		cur_ncrna_region->start = ncRNAs->at(i_reg)->start;
		cur_ncrna_region->end = ncRNAs->at(i_reg)->end;
		cur_ncrna_region->chrom = new char[10];
		strcpy(cur_ncrna_region->chrom, chromosomes[ncRNAs->at(i_reg)->chr_id]);

		ncRNA_regions->push_back(cur_ncrna_region);
	} // i_reg loop.

	/*
	Must make sure that there are no introns in the ncRNA annotations, this is necessary to ensure that the windows that are 
	generated from ncRNA annotations do not contain regions that are low in RNA expression data.

	For this, it is necessary to load the exons for each ncRNA gene and check if any ncRNA gene is divided into two.
	*/
	vector<t_gencode_entry*>* exons = this->get_gtf_entries_by_feature("exon");

//	vector<int>* ncrna_biotypes = new vector<int>();
//
//	/*
//Mt_rRNA,
//Mt_tRNA,
//miRNA,
//misc_RNA,
//rRNA,
//snRNA,
//SnoRNA,
//	*/
//	ncrna_biotypes->push_back(Mt_rRNA);
//	ncrna_biotypes->push_back(Mt_tRNA);
//	ncrna_biotypes->push_back(miRNA);
//	ncrna_biotypes->push_back(misc_RNA);
//	ncrna_biotypes->push_back(rRNA);
//	ncrna_biotypes->push_back(snRNA);
//	ncrna_biotypes->push_back(SnoRNA);
	// Following checks if two ncRNA exons have the same ids.
	vector<t_gencode_entry*>* ncRNA_exons = this->get_gtf_entries_by_gene_biotypes(exons, ncrna_biotypes);

	for(int i_ex = 0; i_ex < ncRNA_exons->size(); i_ex++)
	{
		for(int i_ex2 = i_ex+1; i_ex2 < ncRNA_exons->size(); i_ex2++)
		{
			if(t_string::compare_strings_ci(ncRNA_exons->at(i_ex)->gene_id, ncRNA_exons->at(i_ex2)->gene_id))
			{
				printf("Different exons: <%d-%d>, <%d-%d>\n", ncRNA_exons->at(i_ex)->start, ncRNA_exons->at(i_ex)->end,
					ncRNA_exons->at(i_ex2)->start, ncRNA_exons->at(i_ex2)->end);

				getc(stdin);
			}
		} // i_ex2 loop.
	} // i_ex loop.

	return(ncRNA_regions);
}

vector<t_annot_region*>* t_Gencode_GTF::get_protein_coding_regions()
{
	// Do filtering on the correct biotypes.
	vector<t_annot_region*>* protein_coding_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<t_gencode_entry*>* genes = this->get_gtf_entries_by_feature("gene");

	vector<int>* protein_coding_biotypes = new vector<int>();
	protein_coding_biotypes->push_back(protein_coding);

	vector<t_gencode_entry*>* protein_coding_entries = this->get_gtf_entries_by_gene_biotypes(genes, protein_coding_biotypes);

	printf("%d protein-coding entries.\n", protein_coding_entries->size());

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < protein_coding_entries->size(); i_reg++)
	{
		t_annot_region* cur_coding_region = new t_annot_region();
		cur_coding_region->strand = protein_coding_entries->at(i_reg)->strand;
		cur_coding_region->start = protein_coding_entries->at(i_reg)->start;
		cur_coding_region->end = protein_coding_entries->at(i_reg)->end;
		cur_coding_region->chrom = new char[10];
		strcpy(cur_coding_region->chrom, chromosomes[protein_coding_entries->at(i_reg)->chr_id]);

		protein_coding_regions->push_back(cur_coding_region);
	} // i_reg loop.

	return(protein_coding_regions);
}

vector<t_annot_region*>* t_Gencode_GTF::get_ncRNA_regions()
{
	// Do filtering on the correct biotypes.
	vector<t_annot_region*>* ncRNA_regions = new vector<t_annot_region*>();

	// Filter the entries with respect to the genes.
	vector<t_gencode_entry*>* genes = this->get_gtf_entries_by_feature("gene");

	vector<int>* ncrna_biotypes = new vector<int>();

	/*
Mt_rRNA,
Mt_tRNA,
miRNA,
misc_RNA,
rRNA,
snRNA,
SnoRNA,
	*/
	ncrna_biotypes->push_back(Mt_rRNA);
	ncrna_biotypes->push_back(Mt_tRNA);
	ncrna_biotypes->push_back(tRNAscan);
	ncrna_biotypes->push_back(miRNA);
	ncrna_biotypes->push_back(misc_RNA);
	ncrna_biotypes->push_back(rRNA);
	ncrna_biotypes->push_back(snRNA);
	ncrna_biotypes->push_back(SnoRNA);
	vector<t_gencode_entry*>* ncRNAs = this->get_gtf_entries_by_gene_biotypes(genes, ncrna_biotypes);

	//t_Gencode_GTF::prune_entries_by_start_end(genes);

	printf("%d ncRNA entries.\n", ncRNAs->size());

	// Go over all the regions and add them.
	for(int i_reg = 0; i_reg < ncRNAs->size(); i_reg++)
	{
		t_annot_region* cur_ncrna_region = new t_annot_region();
		cur_ncrna_region->strand = ncRNAs->at(i_reg)->strand;
		cur_ncrna_region->start = ncRNAs->at(i_reg)->start;
		cur_ncrna_region->end = ncRNAs->at(i_reg)->end;
		cur_ncrna_region->chrom = new char[10];
		strcpy(cur_ncrna_region->chrom, chromosomes[ncRNAs->at(i_reg)->chr_id]);

		ncRNA_regions->push_back(cur_ncrna_region);
	} // i_reg loop.

	/*
	Must make sure that there are no introns in the ncRNA annotations, this is necessary to ensure that the windows that are 
	generated from ncRNA annotations do not contain regions that are low in RNA expression data.

	For this, it is necessary to load the exons for each ncRNA gene and check if any ncRNA gene is divided into two.
	*/
	vector<t_gencode_entry*>* exons = this->get_gtf_entries_by_feature("exon");

//	vector<int>* ncrna_biotypes = new vector<int>();
//
//	/*
//Mt_rRNA,
//Mt_tRNA,
//miRNA,
//misc_RNA,
//rRNA,
//snRNA,
//SnoRNA,
//	*/
//	ncrna_biotypes->push_back(Mt_rRNA);
//	ncrna_biotypes->push_back(Mt_tRNA);
//	ncrna_biotypes->push_back(miRNA);
//	ncrna_biotypes->push_back(misc_RNA);
//	ncrna_biotypes->push_back(rRNA);
//	ncrna_biotypes->push_back(snRNA);
//	ncrna_biotypes->push_back(SnoRNA);
	// Following checks if two ncRNA exons have the same ids.
	vector<t_gencode_entry*>* ncRNA_exons = this->get_gtf_entries_by_gene_biotypes(exons, ncrna_biotypes);

	for(int i_ex = 0; i_ex < ncRNA_exons->size(); i_ex++)
	{
		for(int i_ex2 = i_ex+1; i_ex2 < ncRNA_exons->size(); i_ex2++)
		{
			if(t_string::compare_strings_ci(ncRNA_exons->at(i_ex)->gene_id, ncRNA_exons->at(i_ex2)->gene_id))
			{
				printf("Different exons: <%d-%d>, <%d-%d>\n", ncRNA_exons->at(i_ex)->start, ncRNA_exons->at(i_ex)->end,
					ncRNA_exons->at(i_ex2)->start, ncRNA_exons->at(i_ex2)->end);

				getc(stdin);
			}
		} // i_ex2 loop.
	} // i_ex loop.


	return(ncRNA_regions);
}

// Get ncRNA lists.
vector<t_annot_region*>* get_ncRNA_genes()
{
	return(NULL);
}

vector<t_annot_region*>* get_ncRNA_exons()
{
	return(NULL);
}

vector<t_annot_region*>* get_ncRNA_introns()
{
	return(NULL);
}

bool t_Gencode_GTF::sort_genes_by_coord(t_gencode_entry* entry1, t_gencode_entry* entry2)
{
	return(entry1->start < entry2->start);
}




