#ifndef __GENCODE_GTF__
#define __GENCODE_GTF__

#include <vector>

using namespace std;

struct t_annot_region;

char resolve_status(char* status);
char resolve_chromosome(char* chrom);
char resolve_feature(char* feature);
char resolve_biotype(char* type);
char resolve_source(char* source);
char resolve_status(char* source);

void parse_GENCODE_gff_grp_strs(vector<t_annot_region*>* all_regions);

enum{CHR1,
	CHR2,
	CHR3,
	CHR4,
	CHR5,
	CHR6,
	CHR7,
	CHR8,
	CHR9,
	CHR10,
	CHR11,
	CHR12,
	CHR13,
	CHR14,
	CHR15,
	CHR16,
	CHR17,
	CHR18,
	CHR19,
	CHR20,
	CHR21,
	CHR22,
	CHRX,
	CHRY,
	CHRM,
	N_CHRS};

static char chromosomes[][10] = {"chr1",
"chr2",
"chr3",
"chr4",
"chr5",
"chr6",
"chr7",
"chr8",
"chr9",
"chr10",
"chr11",
"chr12",
"chr13",
"chr14",
"chr15",
"chr16",
"chr17",
"chr18",
"chr19",
"chr20",
"chr21",
"chr22",
"chrX",
"chrY",
"chrM"};

enum{ENSEMBL,
	HAVANA,
	N_SOURCES};

static char sources[2][10] = {"ENSEMBL", "HAVANA"};

enum{gene,
transcript,
exon,
UTR,
CDS,
start_codon,
stop_codon,
selenocysteine,
N_FEATURES};

static char features[][20] = {"gene",
"transcript",
"exon",
"UTR",
"CDS",
"start_codon",
"stop_codon",
"selenocysteine"};

enum{IG_C_gene,
IG_D_gene,
IG_J_gene,
IG_V_gene,
TR_C_gene,
TR_J_gene,
TR_V_gene,
IG_pseudogene, 
IG_C_pseudogene,
IG_V_pseudogene,
IG_J_pseudogene,
TR_V_pseudogene,
Mt_rRNA,
Mt_tRNA,
miRNA,
misc_RNA,
rRNA,
snRNA,
SnoRNA,
ncrna_host,
Mt_tRNA_pseudogene,
tRNA_pseudogene,
snoRNA_pseudogene,
snRNA_pseudogene,
scRNA_pseudogene,
rRNA_pseudogene,
misc_RNA_pseudogene,
miRNA_pseudogene,
TR_pseudogene,
TEC,
nonsense_mediated_decay,
retained_intron,
protein_coding,
processed_transcript,
non_coding,
ambiguous_orf, 
antisense,
pseudogene,
processed_pseudogene,
polymorphic_pseudogene, 
retrotransposed,
transcribed_processed_pseudogene,
transcribed_unprocessed_pseudogene,
unitary_pseudogene,
unprocessed_pseudogene, 
artifact,
lincRNA,
LRG_gene,
tRNAscan,
N_BIOTYPES};

static char biotypes[][100]= {"IG_C_gene",
"IG_D_gene",
"IG_J_gene",
"IG_V_gene",
"TR_C_gene",
"TR_J_gene",
"TR_V_gene",
"IG_pseudogene", 
"IG_C_pseudogene",
"IG_V_pseudogene",
"IG_J_pseudogene",
"TR_V_pseudogene",
"Mt_rRNA",
"Mt_tRNA",
"miRNA",
"misc_RNA",
"rRNA",
"snRNA",
"SnoRNA",
"ncrna_host",
"Mt_tRNA_pseudogene",
"tRNA_pseudogene",
"snoRNA_pseudogene",
"snRNA_pseudogene",
"scRNA_pseudogene",
"rRNA_pseudogene",
"misc_RNA_pseudogene",
"miRNA_pseudogene",
"TR_pseudogene",
"TEC",
"nonsense_mediated_decay",
"retained_intron",
"protein_coding",
"processed_transcript",
"non_coding",
"ambiguous_orf", 
"antisense",
"pseudogene",
"processed_pseudogene",
"polymorphic_pseudogene", 
"retrotransposed",
"transcribed_processed_pseudogene",
"transcribed_unprocessed_pseudogene",
"unitary_pseudogene",
"unprocessed_pseudogene", 
"artifact",
"lincRNA",
"LRG_gene",
"tRNAscan"// This is the tRNA gene id from tRNAscan annotations.
};

enum{KNOWN, 
_NULL, 
NOVEL, 
UNKNOWN,
PUTATIVE,
N_STATUS};

static char statuses[][20] = {"KNOWN", 
"NULL", 
"NOVEL", 
"PUTATIVE",
"UNKNOWN"};

struct t_gencode_entry
{
	char chr_id; // From enumerated id's.
	char feature_id; // From enumerated id's.
	char source_id; // From enumerated id's.
	int start;
	int end;
	char strand;

	char gene_type; // From enumerated biotypes.
	char* gene_id;
	char* gene_name;
	char gene_status;
	char* transcript_id;
	char transcript_type; // From enumerated biotypes.
	char* transcript_name;
	char transcript_status;

	char status;
};

/*
Class manages gencode gtf.
*/
class t_Gencode_GTF
{
public:
	t_Gencode_GTF(char* gencode_gtf_file);
	~t_Gencode_GTF();

	// All the entries.
	vector<t_gencode_entry*>* entries;

	// Cluster all the entries with respect to the human gene model

	// Determine the introns from the clustered set: Exclude the pseudogene introns?
	// How to treat the different sets of introns in the isoforms: Take the largest set for each gene.
	vector<t_annot_region*>* get_introns();

	// Get the regions between all the genes.
	vector<t_annot_region*>* get_intergenic_regions();
	vector<t_annot_region*>* get_CDS_regions();
	vector<t_annot_region*>* get_UTR_regions();
	vector<t_annot_region*>* get_ncRNA_regions();
	vector<t_annot_region*>* get_processed_transcript_regions();

	// Get different types of non-coding RNAs.
	vector<t_annot_region*>* get_ncRNA_regions_by_type(int ncrna_type);
	vector<t_annot_region*>* get_protein_coding_regions();

	void prune_entries_by_start_end(vector<t_gencode_entry*>* entries);

	vector<t_gencode_entry*>* get_gtf_entries_by_feature(char* feature);
	static vector<t_gencode_entry*>* get_gtf_entries_by_feature(vector<t_gencode_entry*>* entries, char* feature);
	static vector<t_gencode_entry*>* get_gtf_entries_by_strand(vector<t_gencode_entry*>* entries, char strand);
	static vector<t_gencode_entry*>* get_gtf_entries_by_chr_id(vector<t_gencode_entry*>* entries, int chr_id);
	static vector<t_gencode_entry*>* get_gtf_entries_by_gene_biotypes(vector<t_gencode_entry*>* entries, vector<int>* biotypes);
	static vector<t_gencode_entry*>* get_gtf_entries_by_transcript_biotypes(vector<t_gencode_entry*>* entries, vector<int>* biotypes);

	static vector<t_annot_region*>* get_gtf_annot_regions_by_strand(vector<t_gencode_entry*>* entries, char strand);

	static bool sort_genes_by_coord(t_gencode_entry* entry1, t_gencode_entry* entry2);

	static vector<t_annot_region*>* get_regions_per_encode_entries(vector<t_gencode_entry*>* entries);
};

#endif // __GENCODE_GTF__

