#ifndef __GFF_UTILS__
#define __GFF_UTILS__

struct t_annot_region;

vector<t_annot_region*>* load_GFF(char* gff_fp);
vector<t_annot_region*>* load_GFF_with_line_information(char* gff_fp);

vector<t_annot_region*>* get_gff_regions_per_feature_type(vector<t_annot_region*>* all_regions,
															char* feature_type);

vector<t_annot_region*>* get_gff_regions_per_source_string(vector<t_annot_region*>* all_regions,
															char* source_str);

//vector<t_annot_region*>* get_gff_feature_type(t_annot_region* region);

//char* get_gff_property_per_grp_str(t_annot_region* entry, char* (*get_prop_per_group_str)(char*));

struct t_gff_info
{
	char* feature_type; // Type of the feature: exon, TSS, gene, transcripts, start_codon, ...
	char* group_str; // The information that identifies the entry.
	char* source_str; // This is the experiment id for this entry.
	vector<char*>* prop_ids;
	vector<char*>* prop_vals;
};

// Extract the properties for each entry in the GFF entry list. The callback extracts the property string. This function forms the interface between
// the external functions and the GFF property entries, i.e., group_str entry.
vector<char*>* get_gff_string_properties_per_entry(vector<t_annot_region*>* all_regions,
												char* prop_id);

char* get_prop_val_per_prop_id(t_annot_region* region, char* prop_id);

void parse_group_strs(vector<t_annot_region*>* all_regions, void (*parse_grp_str)(t_annot_region*, vector<char*>*, vector<char*>*));

// Go over all the regions, get the regions with the same property together, choose the largest region that contains the all regions with the selected property.
void form_largest_regions_per_property_type(vector<t_annot_region*>* all_regions,
											vector<t_annot_region*>* merged_selected_regions,
											vector<char*>* sel_region_props, 
											char* prop_id);

void associate_gff_children_2_parents(vector<t_annot_region*>* parent_regions,
	vector<t_annot_region*>* children_regions,
	vector<char*>* (*get_parent_feature_id_per_child_group_str)(t_annot_region*), // Use this to extract the id for the parent using the group string for child.
	char* (*get_parent_feature_id_per_parent_group_str)(t_annot_region*)); // Use this to extract the id for the parent using the group string for parent.

t_annot_region* copy_gff_entry(t_annot_region* gff_entry);

#endif // __GFF_UTILS__

